@@ -6,14 +6,36 @@ from typing import Dict, List
 from .base_service import LLM_Service
 
 
-class LLM_Claude3_Bedrock(LLM_Service):
-    def __init__(self, bedrock_client, model_size):
+class LLM_Claude_Bedrock(LLM_Service):
+    def __init__(self, bedrock_client, model_size, use_caching=True):
         """Constructor
         Arguments:
             bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                 to use when making calls to bedrock models
+            model_size - type of model
+            use_caching - optimize pricing and speed by using caching for prompt and tools
         """
-        if model_size == "Opus":
+        # Only 2 Claude models from Bedrock support caching at the moment
+        if model_size in ["Sonnet 3.7", "Haiku 3.5"]:
+            self.use_caching = use_caching
+        else:
+            self.use_caching = False
+
+        if model_size == "Sonnet 4":
+            self.model_id = "us.anthropic.claude-sonnet-4-20250514-v1:0"
+            self.llm_description = (
+                "Anthropic Claude 4 Sonnet from AWS Bedrock (Medium-size LLM)"
+            )
+            self.price_per_M_input_tokens = 3
+            self.price_per_M_output_tokens = 15
+        elif model_size == "Opus 4":
+            self.model_id = "us.anthropic.claude-opus-4-20250514-v1:0"
+            self.llm_description = (
+                "Anthropic Claude 4 Opus from AWS Bedrock (Large-size LLM)"
+            )
+            self.price_per_M_input_tokens = 15
+            self.price_per_M_output_tokens = 75
+        elif model_size == "Opus":
             self.model_id = "anthropic.claude-3-opus-20240229-v1:0"
             self.llm_description = (
                 "Anthropic Claude 3.0 Opus from AWS Bedrock (Large LLM)"
@@ -91,6 +113,7 @@ class LLM_Claude3_Bedrock(LLM_Service):
         tools=None,
         tool_invoker_fn=None,
         max_retries=25,
+        cur_fail_sleep=60,
     ):
         """
         Invokes the Claude 3 model to run an inference
@@ -123,24 +146,39 @@ class LLM_Claude3_Bedrock(LLM_Service):
         body["system"] = prompt["system"]
         body["messages"] = prompt["messages"].copy()
 
+        # apply caching to System Prompt
+        if self.use_caching:
+            if isinstance(body["system"], str):
+                body["system"] = [
+                    {
+                        "type": "text",
+                        "text": body["system"],
+                        "cache_control": {"type": "ephemeral"},
+                    }
+                ]
+            elif isinstance(body["system"], list):
+                body["system"][-1]["cache_control"] = {"type": "ephemeral"}
+
         if tools is None:
             body["messages"].append({"role": "assistant", "content": postpend})
         else:
+            if self.use_caching:
+                # if caching, append the caching structure to the last tool
+                tools[-1]["cache_control"] = {"type": "ephemeral"}
             body["tools"] = tools
             assert postpend == "", "When using tools, postpend is not supported"
             assert (
                 tool_invoker_fn is not None
             ), "When using tools, a tool invoker must be provided"
 
-        cur_fail_sleep = 60
         for k in range(max_retries):
             try:
-                # print(body)
                 llm_body_changed = True
                 while llm_body_changed:
                     llm_body_changed = False
                     response = self.bedrock_client.invoke_model_with_response_stream(
-                        modelId=self.model_id, body=json.dumps(body)
+                        modelId=self.model_id,
+                        body=json.dumps(body),
                     )
                     word_count = len(re.findall(r"\w+", str(body["messages"])))
                     print(f"Invoking {self.llm_description}. Word count: {word_count}")
@@ -153,9 +191,6 @@ class LLM_Claude3_Bedrock(LLM_Service):
                     cur_ans = x
 
                     if self.cur_tool_spec is not None:
-                        # tool use has been required. Let's do it
-                        # TODO: update upstream to reflect the inclusion of a response
-                        # TODO: probably rework gradio UI to re-instantiate things every chat, or keep an instance per chat ID
                         tool_ans = tool_invoker_fn(
                             self.cur_tool_spec["name"],
                             return_results_only=True,
@@ -206,24 +241,19 @@ class LLM_Claude3_Bedrock(LLM_Service):
                     )
                 return
             except Exception as e:
-                print(
-                    f'Error {str(e)}. Prompt length: {len(str(body["messages"]))}\n\nRetrying {k}...'
-                )
+                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                 time.sleep(int(cur_fail_sleep))
                 cur_fail_sleep *= 1.2
-
-        raise
-        # return response
-
-        # except ClientError:
-        #    logger.error("Couldn't invoke Claude")
-        #    raise
+        yield "Could not invoke the AI model."
 
     def _response_gen(self, response_body, postpend=""):
         cur_ans = ""
         cur_tool_spec = None
         for x in response_body:
             out_dict = json.loads(x["chunk"]["bytes"])
+            if "type" in out_dict.keys() and out_dict["type"] == "message_start":
+                print(x)
+
             txt = ""
             if "content_block" in out_dict.keys():
                 if out_dict["content_block"]["type"] == "text":
@@ -300,6 +330,7 @@ class LLM_Mistral_Bedrock(LLM_Service):
         max_retries=25,
         tools=None,
         tool_invoker_fn=None,
+        cur_fail_sleep=6,
     ):
         """
         Invokes the Llama2 large-language model to run an inference
@@ -321,61 +352,53 @@ class LLM_Mistral_Bedrock(LLM_Service):
         # body['stop_sequences'] = body['stop_sequences'] + extra_stop_sequences
         body["prompt"] = prompt + postpend
 
-        cur_fail_sleep = 1
         for k in range(max_retries):
-            # try:
-            response = self.bedrock_client.invoke_model_with_response_stream(
-                modelId=self.model_id, body=json.dumps(body)
-            )
-            word_count = len(re.findall(r"\w+", body["prompt"]))
-            print(f"Invoking {self.llm_description}. Word count: {word_count}")
+            try:
+                response = self.bedrock_client.invoke_model_with_response_stream(
+                    modelId=self.model_id, body=json.dumps(body)
+                )
+                word_count = len(re.findall(r"\w+", body["prompt"]))
+                print(f"Invoking {self.llm_description}. Word count: {word_count}")
 
-            cur_ans = ""
-            for x in response["body"]:
-                partial = json.loads(x["chunk"]["bytes"])["outputs"][0]["text"]
-                cur_ans += partial
+                cur_ans = ""
+                for x in response["body"]:
+                    partial = json.loads(x["chunk"]["bytes"])["outputs"][0]["text"]
+                    cur_ans += partial
 
-                # we include those again so there's no need to leave these in the answer
-                cur_ans = cur_ans.replace("<s>", "").replace("</s>", "")
+                    # we include those again so there's no need to leave these in the answer
+                    cur_ans = cur_ans.replace("<s>", "").replace("</s>", "")
 
-                # we have to manually check for stopword generation
-                extra_stop_break = False
-                for x in extra_stop_sequences:
-                    cur_split = cur_ans.split(x)
-                    if len(cur_split) > 1:
-                        cur_ans = cur_split[0] + x
-                        extra_stop_break = True
+                    # we have to manually check for stopword generation
+                    extra_stop_break = False
+                    for x in extra_stop_sequences:
+                        cur_split = cur_ans.split(x)
+                        if len(cur_split) > 1:
+                            cur_ans = cur_split[0] + x
+                            extra_stop_break = True
+                            break
+
+                    yield postpend + cur_ans
+
+                    if extra_stop_break:
                         break
+                    # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']
 
-                yield postpend + cur_ans
-
-                if extra_stop_break:
-                    break
-                # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']
-
-            ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
-            self.word_counts.append(
-                {
-                    "request_word_count": word_count,
-                    "answer_word_count": ans_word_count,
-                    "price_estimate": 0.001
-                    * (0.00195 * word_count + 0.00256 * ans_word_count),
-                    "exec_time_in_s": time.time() - t0,
-                }
-            )
-
-            return
-        # except Exception as e:
-        #    print(f'Error {str(e)}. Prompt length: {len(body["prompt"])}\n\nRetrying {k}...')
-        #    time.sleep(int(cur_fail_sleep))
-        #    cur_fail_sleep *= 1.2
-
-        raise
-        # return response
-
-        # except ClientError:
-        #    logger.error("Couldn't invoke Claude")
-        #    raise
+                ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
+                self.word_counts.append(
+                    {
+                        "request_word_count": word_count,
+                        "answer_word_count": ans_word_count,
+                        "price_estimate": 0.001
+                        * (0.00195 * word_count + 0.00256 * ans_word_count),
+                        "exec_time_in_s": time.time() - t0,
+                    }
+                )
+                return
+            except Exception as e:
+                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
+                time.sleep(int(cur_fail_sleep))
+                cur_fail_sleep *= 1.2
+        yield "Could not invoke the AI model."
 
 
 class LLM_Claude2_1_Bedrock(LLM_Service):
@@ -410,7 +433,12 @@ class LLM_Claude2_1_Bedrock(LLM_Service):
         return format_messages_for_claude(msg_list)
 
     def invoke_streaming(
-        self, prompt, postpend="", extra_stop_sequences=[], max_retries=25
+        self,
+        prompt,
+        postpend="",
+        extra_stop_sequences=[],
+        max_retries=25,
+        cur_fail_sleep=6,
     ):
         """
         Invokes the Claude large-language model to run an inference
@@ -432,7 +460,6 @@ class LLM_Claude2_1_Bedrock(LLM_Service):
         body["stop_sequences"] = body["stop_sequences"] + extra_stop_sequences
         body["prompt"] = prompt + postpend
 
-        cur_fail_sleep = 1
         for k in range(max_retries):
             try:
                 response = self.bedrock_client.invoke_model_with_response_stream(
@@ -465,21 +492,12 @@ class LLM_Claude2_1_Bedrock(LLM_Service):
                         "exec_time_in_s": time.time() - t0,
                     }
                 )
-
                 return
             except Exception as e:
-                print(
-                    f'Error {str(e)}. Prompt length: {len(body["prompt"])}\n\nRetrying {k}...'
-                )
+                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                 time.sleep(int(cur_fail_sleep))
                 cur_fail_sleep *= 1.2
-
-        raise
-        # return response
-
-        # except ClientError:
-        #    logger.error("Couldn't invoke Claude")
-        #    raise
+        yield "Could not invoke the AI model."
 
 
 class LLM_Claude_Instant_1_2_Bedrock(LLM_Service):
@@ -513,7 +531,12 @@ class LLM_Claude_Instant_1_2_Bedrock(LLM_Service):
         return format_messages_for_claude(msg_list)
 
     def invoke_streaming(
-        self, prompt, postpend="", extra_stop_sequences=[], max_retries=25
+        self,
+        prompt,
+        postpend="",
+        extra_stop_sequences=[],
+        max_retries=25,
+        cur_fail_sleep=1,
     ):
         """
         Invokes the Claude large-language model to run an inference
@@ -535,7 +558,6 @@ class LLM_Claude_Instant_1_2_Bedrock(LLM_Service):
         body["stop_sequences"] = body["stop_sequences"] + extra_stop_sequences
         body["prompt"] = prompt + postpend
 
-        cur_fail_sleep = 1
         for k in range(max_retries):
             try:
                 response = self.bedrock_client.invoke_model_with_response_stream(
@@ -568,21 +590,12 @@ class LLM_Claude_Instant_1_2_Bedrock(LLM_Service):
                         "exec_time_in_s": time.time() - t0,
                     }
                 )
-
                 return
             except Exception as e:
-                print(
-                    f'Error {str(e)}. Prompt length: {len(body["prompt"])}\n\nRetrying {k}...'
-                )
+                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
                 time.sleep(int(cur_fail_sleep))
                 cur_fail_sleep *= 1.2
-
-        raise
-        # return response
-
-        # except ClientError:
-        #    logger.error("Couldn't invoke Claude")
-        #    raise
+        yield "Could not invoke the AI model."
 
 
 class LLM_Llama13b(LLM_Service):
@@ -616,7 +629,12 @@ class LLM_Llama13b(LLM_Service):
         return format_messages_for_llama(msg_list)
 
     def invoke_streaming(
-        self, prompt, postpend="", extra_stop_sequences=[], max_retries=25
+        self,
+        prompt,
+        postpend="",
+        extra_stop_sequences=[],
+        max_retries=25,
+        cur_fail_sleep=1,
     ):
         """
         Invokes the Llama2 large-language model to run an inference
@@ -638,61 +656,53 @@ class LLM_Llama13b(LLM_Service):
         # body['stop_sequences'] = body['stop_sequences'] + extra_stop_sequences
         body["prompt"] = prompt + postpend
 
-        cur_fail_sleep = 1
         for k in range(max_retries):
-            # try:
-            response = self.bedrock_client.invoke_model_with_response_stream(
-                modelId="meta.llama2-13b-chat-v1", body=json.dumps(body)
-            )
-            word_count = len(re.findall(r"\w+", body["prompt"]))
-            print(f"Invoking Llama 2 chat 13b. Word count: {word_count}")
+            try:
+                response = self.bedrock_client.invoke_model_with_response_stream(
+                    modelId="meta.llama2-13b-chat-v1", body=json.dumps(body)
+                )
+                word_count = len(re.findall(r"\w+", body["prompt"]))
+                print(f"Invoking Llama 2 chat 13b. Word count: {word_count}")
 
-            cur_ans = ""
-            for x in response["body"]:
-                partial = json.loads(x["chunk"]["bytes"])["generation"]
-                cur_ans += partial
+                cur_ans = ""
+                for x in response["body"]:
+                    partial = json.loads(x["chunk"]["bytes"])["generation"]
+                    cur_ans += partial
 
-                # we include those again so there's no need to leave these in the answer
-                cur_ans = cur_ans.replace("<s>", "").replace("</s>", "")
+                    # we include those again so there's no need to leave these in the answer
+                    cur_ans = cur_ans.replace("<s>", "").replace("</s>", "")
 
-                # we have to manually check for stopword generation
-                extra_stop_break = False
-                for x in extra_stop_sequences:
-                    cur_split = cur_ans.split(x)
-                    if len(cur_split) > 1:
-                        cur_ans = cur_split[0] + x
-                        extra_stop_break = True
+                    # we have to manually check for stopword generation
+                    extra_stop_break = False
+                    for x in extra_stop_sequences:
+                        cur_split = cur_ans.split(x)
+                        if len(cur_split) > 1:
+                            cur_ans = cur_split[0] + x
+                            extra_stop_break = True
+                            break
+
+                    yield postpend + cur_ans
+
+                    if extra_stop_break:
                         break
+                    # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']
 
-                yield postpend + cur_ans
-
-                if extra_stop_break:
-                    break
-                # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']
-
-            ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
-            self.word_counts.append(
-                {
-                    "request_word_count": word_count,
-                    "answer_word_count": ans_word_count,
-                    "price_estimate": 0.001
-                    * (0.00075 * word_count + 0.001 * ans_word_count),
-                    "exec_time_in_s": time.time() - t0,
-                }
-            )
-
-            return
-        # except Exception as e:
-        #    print(f'Error {str(e)}. Prompt length: {len(body["prompt"])}\n\nRetrying {k}...')
-        #    time.sleep(int(cur_fail_sleep))
-        #    cur_fail_sleep *= 1.2
-
-        raise
-        # return response
-
-        # except ClientError:
-        #    logger.error("Couldn't invoke Claude")
-        #    raise
+                ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
+                self.word_counts.append(
+                    {
+                        "request_word_count": word_count,
+                        "answer_word_count": ans_word_count,
+                        "price_estimate": 0.001
+                        * (0.00075 * word_count + 0.001 * ans_word_count),
+                        "exec_time_in_s": time.time() - t0,
+                    }
+                )
+                return
+            except Exception as e:
+                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
+                time.sleep(int(cur_fail_sleep))
+                cur_fail_sleep *= 1.2
+        yield "Could not invoke the AI model."
 
 
 class LLM_Llama70b(LLM_Service):
@@ -725,7 +735,12 @@ class LLM_Llama70b(LLM_Service):
         return format_messages_for_llama(msg_list)
 
     def invoke_streaming(
-        self, prompt, postpend="", extra_stop_sequences=[], max_retries=25
+        self,
+        prompt,
+        postpend="",
+        extra_stop_sequences=[],
+        max_retries=25,
+        cur_fail_sleep=1,
     ):
         """
         Invokes the Llama2 large-language model to run an inference
@@ -747,61 +762,53 @@ class LLM_Llama70b(LLM_Service):
         # body['stop_sequences'] = body['stop_sequences'] + extra_stop_sequences
         body["prompt"] = prompt + postpend
 
-        cur_fail_sleep = 1
         for k in range(max_retries):
-            # try:
-            response = self.bedrock_client.invoke_model_with_response_stream(
-                modelId="meta.llama2-70b-chat-v1", body=json.dumps(body)
-            )
-            word_count = len(re.findall(r"\w+", body["prompt"]))
-            print(f"Invoking Llama 2 chat 70b. Word count: {word_count}")
+            try:
+                response = self.bedrock_client.invoke_model_with_response_stream(
+                    modelId="meta.llama2-70b-chat-v1", body=json.dumps(body)
+                )
+                word_count = len(re.findall(r"\w+", body["prompt"]))
+                print(f"Invoking Llama 2 chat 70b. Word count: {word_count}")
 
-            cur_ans = ""
-            for x in response["body"]:
-                partial = json.loads(x["chunk"]["bytes"])["generation"]
-                cur_ans += partial
+                cur_ans = ""
+                for x in response["body"]:
+                    partial = json.loads(x["chunk"]["bytes"])["generation"]
+                    cur_ans += partial
 
-                # we include those again so there's no need to leave these in the answer
-                cur_ans = cur_ans.replace("<s>", "").replace("</s>", "")
+                    # we include those again so there's no need to leave these in the answer
+                    cur_ans = cur_ans.replace("<s>", "").replace("</s>", "")
 
-                # we have to manually check for stopword generation
-                extra_stop_break = False
-                for x in extra_stop_sequences:
-                    cur_split = cur_ans.split(x)
-                    if len(cur_split) > 1:
-                        cur_ans = cur_split[0] + x
-                        extra_stop_break = True
+                    # we have to manually check for stopword generation
+                    extra_stop_break = False
+                    for x in extra_stop_sequences:
+                        cur_split = cur_ans.split(x)
+                        if len(cur_split) > 1:
+                            cur_ans = cur_split[0] + x
+                            extra_stop_break = True
+                            break
+
+                    yield postpend + cur_ans
+
+                    if extra_stop_break:
                         break
+                    # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']
 
-                yield postpend + cur_ans
-
-                if extra_stop_break:
-                    break
-                # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']
-
-            ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
-            self.word_counts.append(
-                {
-                    "request_word_count": word_count,
-                    "answer_word_count": ans_word_count,
-                    "price_estimate": 0.001
-                    * (0.00195 * word_count + 0.00256 * ans_word_count),
-                    "exec_time_in_s": time.time() - t0,
-                }
-            )
-
-            return
-        # except Exception as e:
-        #    print(f'Error {str(e)}. Prompt length: {len(body["prompt"])}\n\nRetrying {k}...')
-        #    time.sleep(int(cur_fail_sleep))
-        #    cur_fail_sleep *= 1.2
-
-        raise
-        # return response
-
-        # except ClientError:
-        #    logger.error("Couldn't invoke Claude")
-        #    raise
+                ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
+                self.word_counts.append(
+                    {
+                        "request_word_count": word_count,
+                        "answer_word_count": ans_word_count,
+                        "price_estimate": 0.001
+                        * (0.00195 * word_count + 0.00256 * ans_word_count),
+                        "exec_time_in_s": time.time() - t0,
+                    }
+                )
+                return
+            except Exception as e:
+                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
+                time.sleep(int(cur_fail_sleep))
+                cur_fail_sleep *= 1.2
+        yield "Could not invoke the AI model."
 
 
 class LLM_Llama3(LLM_Service):
@@ -881,6 +888,7 @@ class LLM_Llama3(LLM_Service):
         max_retries=25,
         tools=None,
         tool_invoker_fn=None,
+        cur_fail_sleep=1,
     ):
         """
         Invokes the Llama3 large-language model to run an inference
@@ -905,58 +913,50 @@ class LLM_Llama3(LLM_Service):
         # body['stop_sequences'] = body['stop_sequences'] + extra_stop_sequences
         body["prompt"] = prompt + postpend
 
-        cur_fail_sleep = 1
         for k in range(max_retries):
-            # try:
-            response = self.bedrock_client.invoke_model_with_response_stream(
-                modelId=self.model_id, body=json.dumps(body)
-            )
-            word_count = len(re.findall(r"\w+", body["prompt"]))
-            print(f"Invoking {self.llm_description}. Word count: {word_count}")
+            try:
+                response = self.bedrock_client.invoke_model_with_response_stream(
+                    modelId=self.model_id, body=json.dumps(body)
+                )
+                word_count = len(re.findall(r"\w+", body["prompt"]))
+                print(f"Invoking {self.llm_description}. Word count: {word_count}")
 
-            cur_ans = ""
-            for x in response["body"]:
-                partial = json.loads(x["chunk"]["bytes"])["generation"]
-                cur_ans += partial
+                cur_ans = ""
+                for x in response["body"]:
+                    partial = json.loads(x["chunk"]["bytes"])["generation"]
+                    cur_ans += partial
 
-                # we have to manually check for stopword generation
-                extra_stop_break = False
-                for x in extra_stop_sequences:
-                    cur_split = cur_ans.split(x)
-                    if len(cur_split) > 1:
-                        cur_ans = cur_split[0] + x
-                        extra_stop_break = True
+                    # we have to manually check for stopword generation
+                    extra_stop_break = False
+                    for x in extra_stop_sequences:
+                        cur_split = cur_ans.split(x)
+                        if len(cur_split) > 1:
+                            cur_ans = cur_split[0] + x
+                            extra_stop_break = True
+                            break
+
+                    yield postpend + cur_ans
+
+                    if extra_stop_break:
                         break
+                    # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']
 
-                yield postpend + cur_ans
-
-                if extra_stop_break:
-                    break
-                # stop_reason = json.loads(x['chunk']['bytes'])['stop_reason']
-
-            ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
-            self.word_counts.append(
-                {
-                    "request_word_count": word_count,
-                    "answer_word_count": ans_word_count,
-                    "price_estimate": 0.001
-                    * (0.00195 * word_count + 0.00256 * ans_word_count),
-                    "exec_time_in_s": time.time() - t0,
-                }
-            )
-
-            return
-        # except Exception as e:
-        #    print(f'Error {str(e)}. Prompt length: {len(body["prompt"])}\n\nRetrying {k}...')
-        #    time.sleep(int(cur_fail_sleep))
-        #    cur_fail_sleep *= 1.2
-
-        raise
-        # return response
-
-        # except ClientError:
-        #    logger.error("Couldn't invoke Claude")
-        #    raise
+                ans_word_count = len(re.findall(r"\w+", postpend + cur_ans))
+                self.word_counts.append(
+                    {
+                        "request_word_count": word_count,
+                        "answer_word_count": ans_word_count,
+                        "price_estimate": 0.001
+                        * (0.00195 * word_count + 0.00256 * ans_word_count),
+                        "exec_time_in_s": time.time() - t0,
+                    }
+                )
+                return
+            except Exception as e:
+                yield f"Error {str(e)}. Waiting {int(cur_fail_sleep)} s. Retrying {k+1}/{max_retries}..."
+                time.sleep(int(cur_fail_sleep))
+                cur_fail_sleep *= 1.2
+        yield "Could not invoke the AI model."
 
 
 def format_messages_for_claude(messages: List[Dict[str, str]]) -> List[str]:
